require("scripts/game/scene/scene_config")
require("scripts/game/scene/scene_protocal")
require("scripts/game/scene/sceneobj/scene_obj")
require("scripts/game/scene/sceneobj/character")
require("scripts/game/scene/sceneobj/role")
require("scripts/game/scene/sceneobj/main_role")
require("scripts/game/scene/sceneobj/monster")
require("scripts/game/scene/sceneobj/npc")
require("scripts/game/scene/sceneobj/fall_item")
require("scripts/game/scene/sceneobj/effect_obj")
require("scripts/game/scene/sceneobj/special_obj")
require("scripts/game/scene/sceneobj/decoration")
require("scripts/game/scene/sceneobj/hero")
require("scripts/game/scene/sceneobj/gather")
require("scripts/game/scene/sceneobj/totem")
require("scripts/game/scene/sceneobj/human_monster")
require("scripts/game/scene/sceneobj/pet")
require("scripts/game/scene/sceneobj/display_monster")
require("scripts/game/scene/sceneobj/robot")
require("scripts/game/scene/sceneobj/richman")
require("scripts/game/scene/sceneobj/role_ghost")

require("scripts/game/scenelogic/scene_logic")

local fps_title = 45
local fps_wing = 41
local fps_douli = 37
local fps_hero = 34
local fps_pet = 40

Scene = Scene or BaseClass(BaseController)

function Scene:__init()
	if Scene.Instance then
		ErrorLog("[Scene] Attempt to create singleton twice!")
		return
	end
	Scene.Instance = self

	self.scene_config = nil

	self.touch_id = nil

	self.touch_point = nil

	local main_role_vo = GameVoManager.Instance:GetMainRoleVo()
	self.main_role = MainRole.New(main_role_vo)							-- 预创建一个main_role，防止Get到一个nil

	self.obj_list = {}													-- 场景对象列表
	self.obj_group_list = {}											-- 按对象类型分组存储的对象列表
	self.effect_list = {}                                               -- 地图动态特效管理器

	self.scene_area_info_list = {}										-- 场景区域信息
	self.scene_area_req_list = {}										-- 场景区域请求记录

	self.is_in_update = false											-- 是否正在执行Update过程
	self.delay_handle_funcs = {}										-- 延时处理的函数列表(用于处理延时删除等逻辑)

	self.scene_logic = SceneLogic.Create(0, 0)							-- 场景逻辑

	self.is_in_door = nil												-- 是否在传送点内，nil表示刚进场景未初始化

	
	self.last_bag_full_time = 0 										-- 拾取道具背包已满提示
	self.pick_color_flag = false										-- 自动拾取设置

	self.camera_end_scale = 1 											-- 场景象机缩放目标值
	self.camera_scale_speed = 0 										-- 场景象机缩放速度

	self.cur_main_role_logic_pos = {x = 0, y = 0}						-- 主角当前位置

	self.prev_scene_id = 0                                              
	self.prev_fb_id = 0
	self.cur_fb_id = 0

	self.is_check_all_hide = false                                      -- 当前场景是否检测屏蔽区域和人物任务距离隐身
	self.check_all_hide_range = 100                                     -- 检测吃鸡视野
	self.eat_chicken_range_list = nil                                   -- 吃鸡地图区域
	self.client_scene_cfg = nil
	self.area_effect_indexs = nil

	self.is_pingbi_other_role = false									-- 是否屏蔽其它玩家
	self.is_pingbi_guild_role = false									-- 是否屏蔽友方
	self.is_pingbi_skill_effect = false									-- 是否屏蔽他人技能特效
	self.is_pingbi_self_skill_effect = false							-- 是否屏蔽自己技能特效
	self.is_no_title = false 											-- 是否屏蔽称号
	self.is_simple_name = false 											-- 是否屏蔽名字
	self.is_no_fall_name = false										-- 是否屏蔽掉落名字
	self.is_show_monster_name = false									-- 是否显示怪名字
	self.is_pinbi_monster = false										-- 是否屏蔽普通怪
	self.is_pinbi_pet = false											-- 是否屏蔽宠物怪
	self.is_pinbi_wing = false											-- 是否屏蔽翅膀
	self.is_pinbi_douli = false											-- 是否屏蔽斗笠
	self.is_show_pet_hero_name = false                                  -- 是否显示战将，宠物名称
	self.is_pinbi_hero = false                                          -- 是否屏蔽英雄

	self.total_role_count = 0 											-- 场景上的角色数量(不含主角)
	self.limit_appear_role_list = {} 									-- 被限制显示的角色列表(objid 为Key)
	self.max_appear_role_count = SceneAppearRoleCount.Max				-- 同屏限制最大人数

	self:RegisterAllProtocols()											-- 注册所有需要响应的协议
	self:RegisterAllEvents()											-- 注册所有需要监听的事件

	self.has_pick_fallitem_t = {}
	self.has_pick_count = 0
	self.auto_fight_record_not_pick = {}

	self.fps_now = 60
	self.fps_time = 0
	self.fps_tick = 0
	self.total_fps = 0
	self.is_check_fps = false

	self.cur_area_info = {
		area_name = "",
		attr_t = {},
	}

	-- FpsSampleUtil.Instance:SetLowFpsWarnCallback(BindTool.Bind1(self.LowFpsWarnCallback, self))
	-- FpsSampleUtil.Instance:SetCancleLowFpsWarnCallback(BindTool.Bind1(self.CancleLowFpsWarnCallback, self))

	FpsSampleUtil.Instance:SetFpsCallback(BindTool.Bind1(self.OnFpsCallback, self))

	Runner.Instance:AddRunObj(self, 3)

	MapLoading.New()
	self.draw_node = cc.DrawNode:create()
	HandleRenderUnit:GetCoreScene():addChildToRenderGroup(self.draw_node, GRQ_BLOCK)
end

function Scene:__delete()
	self:ClearScene()

	if self.main_role then
		self.main_role:DeleteMe()
		self.main_role = nil
	end

	MapLoading.Instance:DeleteMe()
	Scene.Instance = nil


	if ItemData.Instance then
		ItemData.Instance:UnNotifyDataChangeCallBack(self.itemdata_change_callback)
	end

	if RoleData.Instance then
		RoleData.Instance:UnNotifyAttrChange(self.roledata_change_callback)
	end	

	Runner.Instance:RemoveRunObj(self)
end

--添加场景特效
function Scene:AddEffect(effect_vo)
	local obj = EffectObj.New(effect_vo)
	obj:Init(nil)
	table.insert(self.effect_list,obj)
end	

function Scene:GetFallItemScene() --掉落物品特效表
	if not self.fall_item_config then
		self.fall_item_config = ConfigManager.Instance:GetAutoConfig("fall_item_config").fall_item_list
	end
	return self.fall_item_config
end


function Scene:ClearScene()
	self.area_effect_indexs = nil
	self.check_all_hide_range = 100
	self.scene_config = nil
	self.has_pick_fallitem_t = {}
	self.has_pick_count = 0
	self.main_role:GetModel():SetVisible(false)

	for _, v in pairs(self.obj_list) do
		if v ~= self.main_role then
			v:DeleteMe()
		end
	end
	self.obj_list = {}
	self.obj_group_list = {}

	--清理所有动态特效
	for i = #self.effect_list , 1, -1 do
		local obj = table.remove(self.effect_list,i)
		obj:DeleteMe()
	end	
	self.effect_list = {}

	self.total_role_count = 0
	self.limit_appear_role_list = {}

	
	self.is_in_update = false
	self.delay_handle_funcs = {}

	self.cur_main_role_logic_pos.x = 0
	self.cur_main_role_logic_pos.y = 0

	if nil ~= self.scene_logic then
		self.scene_logic:DeleteMe()
		self.scene_logic = nil
	end

	self.is_in_door = nil

	HandleGameMapHandler:SetCameraScale(1)
	self.camera_end_scale = 1

	-- AudioManager.Instance:StopMusic()
	FpsSampleUtil.Instance:SetFpsSampleInvalid(false)

	self.draw_node:clear()
end



function Scene:RegisterAllEvents()
	self:Bind(LayerEventType.TOUCH_BEGAN, BindTool.Bind1(self.OnTouchBegan, self))
	self:Bind(LayerEventType.TOUCH_MOVED, BindTool.Bind1(self.OnTouchMoved, self))
	self:Bind(LayerEventType.TOUCH_ENDED, BindTool.Bind1(self.OnTouchEnded, self))

	self:Bind(SceneEventType.SCENE_LOADING_STATE_ENTER, BindTool.Bind1(self.ChangeScene, self))

	self:Bind(ObjectEventType.MAIN_ROLE_POS_CHANGE, BindTool.Bind1(self.OnMainRolePosChange, self))
	self:Bind(ObjectEventType.MAIN_ROLE_MOVE_START, BindTool.Bind1(self.OnMainRoleMoveStart, self))
	self:Bind(ObjectEventType.MAIN_ROLE_MOVE_END, BindTool.Bind1(self.OnMainRoleMoveEnd, self))
	self:Bind(ObjectEventType.MAIN_ROLE_DEAD, BindTool.Bind1(self.OnMainRoleDead, self))

	self:Bind(SettingEventType.SYSTEM_SETTING_CHANGE, BindTool.Bind1(self.OnSysSettingChange, self))
	self:Bind(SettingEventType.GUAJI_SETTING_CHANGE, BindTool.Bind1(self.OnGuaJiSettingChange, self))

	self:Bind(ObjectEventType.SCENE_OBJ_POS_CHANGE,BindTool.Bind1(self.OnSceneObjPosChange, self))

	self.itemdata_change_callback = BindTool.Bind1(self.ItemDataChangeCallback, self)
	ItemData.Instance:NotifyDataChangeCallBack(self.itemdata_change_callback)

	self.roledata_change_callback = BindTool.Bind(self.OnRoleAttrChange, self)
	RoleData.Instance:NotifyAttrChange(self.roledata_change_callback)
end

--场景手指点击
function Scene:OnTouchBegan(touch, event)
	local main_role = Scene.Instance:GetMainRole()
	

	if nil == self.touch_id then
		self.touch_id = touch:getId()
		self.touch_point = touch:getLocation()

		if not GuideCtrl.Instance:IsFuncGuideing() and not  MovieGuideCtrl.Instance:IsGuiding() then
			MainuiCtrl.Instance:PushMainuiNavBtnToggle(false)
		end	
		
		local world_pos = HandleRenderUnit:ScreenToWorld(self.touch_point)
		if SkillData.is_select_skill ~= 0 then
			main_role:StopMove()
			if SkillData.Instance:CanUseSkill(SkillData.is_select_skill) then
				local logic_pos = HandleRenderUnit:WorldToLogic(world_pos)
				main_role:SetPrepareSkill(SkillData.is_select_skill,logic_pos.x,logic_pos.y)
			end
		else	
			
			local click_obj = self:GetClickObj(world_pos.x, world_pos.y)
			if not click_obj then
				
				local screen_pos = HandleRenderUnit:WorldToScreen(cc.p(main_role:GetRealPos()))
				local dir = GameMath.GetDirectionNumber(self.touch_point.x - screen_pos.x, self.touch_point.y - screen_pos.y)
				main_role:SetClickDir(dir)
				main_role:SetIsClickMove(true)
				main_role:StopMove()
			end
		end
	end
end

--场景手指移动
function Scene:OnTouchMoved(touch, event)
	if self.touch_id == touch:getId() then
		self.touch_point = touch:getLocation()

		local main_role = Scene.Instance:GetMainRole()
		local screen_pos = HandleRenderUnit:WorldToScreen(cc.p(main_role:GetRealPos()))
		local dir = GameMath.GetDirectionNumber(self.touch_point.x - screen_pos.x, self.touch_point.y - screen_pos.y)
		main_role:SetClickDir(dir)
	end
end

--场景手指松开
function Scene:OnTouchEnded(touch, event)
	local main_role = Scene.Instance:GetMainRole()
	main_role:SetIsClickMove(false)
	

	if self.touch_id ~= touch:getId() then
		return
	end

	self.touch_id = nil
	self.touch_point = nil

	
	local location = touch:getLocation()
	local world_pos = HandleRenderUnit:ScreenToWorld(location)
	local click_obj = self:GetClickObj(world_pos.x, world_pos.y)
	local srcObj = SceneObj.CurrentSelectObj

	if click_obj then
		if srcObj then
			if srcObj == click_obj then
				main_role:DoSelectSceneObj(click_obj,true)
			else	
				if 	click_obj:GetType() ==  SceneObjType.Npc then
					main_role:DoSelectSceneObj(click_obj,true)
				else
					main_role:DoSelectSceneObj(click_obj,false)
				end
			end	
		else
			if 	click_obj:GetType() ==  SceneObjType.Npc then
				main_role:DoSelectSceneObj(click_obj,true)
			else
				main_role:DoSelectSceneObj(click_obj,false)
			end	
		end
	else
		main_role:StopMove()	
	end
		
end

function Scene:GetClickObj(x, y)
	local click_obj = nil
	local click_obj_zorder = -999999
	local zorder = 0

	for _, v in pairs(self.obj_list) do
		if v:GetModel():IsVisible() and v:IsClick(x, y) then
			zorder = v:GetLocalZOrder()

			local obj_type = v:GetType()
			if obj_type == SceneObjType.Npc 
				or obj_type == SceneObjType.FallItem then --npc优先
				zorder = zorder + 20000
			end

			if zorder > click_obj_zorder then
				click_obj = v
				click_obj_zorder = zorder
			end
		end
	end

	return click_obj
end

function Scene:GetSceneConfig()
	return self.scene_config
end

function Scene:GetSceneId()
	return self.scene_config and self.scene_config.id or 0
end

function Scene:GetPrevSceneId()
	return self.prev_scene_id
end	

function Scene:GetSceneName()
	return self.scene_config and self.scene_config.name or ""
end

function Scene:GetSceneType()
	return self.scene_config and self.scene_config.scene_type or 0
end

function Scene:GetCurAreaInfo()
	return self.cur_area_info
end

function Scene:GetSceneLogic()
	return self.scene_logic
end

function Scene:GetFbID()
	return self.cur_fb_id, self.prev_fb_id
end	

function Scene:ChangeScene(scene_id, scene_type, fb_id)
	local temp_scene_id = self:GetSceneId()
	if temp_scene_id ~= scene_id then
		self.prev_scene_id = temp_scene_id
	end

	if self.cur_fb_id ~= fb_id then
		self.prev_fb_id = self.cur_fb_id
		self.cur_fb_id = fb_id
	end	

	local scene_config = ConfigManager.Instance:GetSceneConfig(scene_id)
	if nil == scene_config then
		Log("scene_config not find, scene_id:" .. scene_id)
		return
	end
	local old_scene_type = nil ~= self.scene_config and self.scene_config.scene_type or SceneType.Common
	local new_scene_type = scene_type
	scene_config.scene_type = scene_type
	if self.scene_logic ~= nil and self.scene_logic:GetSceneType() ~= nil then
		self.scene_logic:Out(old_scene_type, new_scene_type)
	end

	self:ClearScene()
	
	self.scene_config = scene_config
	self.scene_logic = SceneLogic.Create(self.scene_config.scene_type, fb_id, scene_id)

	HandleGameMapHandler:ChangeScene(scene_id)
	HandleRenderUnit:UpdateWorldSize()
	HandleGameMapHandler:ResetViewCenterPoint()

	local client_scene_cfg = SceneForClientConfig[scene_id]
	if client_scene_cfg then
		self.is_check_all_hide = client_scene_cfg.isCheckAllHide or false
		self.check_all_hide_range = client_scene_cfg.defaultRange or 100
		self.eat_chicken_range_list = client_scene_cfg.effectRanges
	else
		self.is_check_all_hide = false	
		self.eat_chicken_range_list = nil
	end
	self.client_scene_cfg = client_scene_cfg

	function loadSceneComplete()
		self.scene_logic:Enter(old_scene_type, new_scene_type)

		GlobalEventSystem:Fire(SceneEventType.SCENE_CHANGE_COMPLETE, old_scene_type, new_scene_type)

		local scene_audio_cfg = ConfigManager.Instance:GetAutoConfig("audio_auto").scene[self:GetSceneId()]
		if nil ~= scene_audio_cfg then
			local audio_id = scene_audio_cfg.audio_id
			if audio_id then
				AudioManager.Instance:PlayMusic(ResPath.GetAudioBgResPath(audio_id))
			end	
		end
		
		FpsSampleUtil.Instance:SetFpsSampleInvalid(true)

		HandleGameMapHandler:OnLoadingSceneQuit()
		self:CheckClientObj()

		self.main_role:OnSceneChange()

		if old_scene_type == SceneType.Fubun or  old_scene_type == SceneType.Activity then
			self.main_role:SetIsAutoFight(false)
		end	
		if self.client_scene_cfg and self.client_scene_cfg.checkItems then
			self:ItemDataChangeCallback(nil,self.client_scene_cfg.checkItems[1])
		end	

		self.is_check_fps = true
	end

	-- if Story.Instance:GetIsOpeningAnimatino() then --在开场动画中创建的场景是虚拟场景，不请求数据
	-- 	loadSceneComplete()
	-- 	return
	-- end

	self:CreateMainRole()
	self.main_role:SetIsLoading(true)
	
	self:InitClientObj()

	--作假进度条
	MapLoading.Instance:SetLoadCompleteCallBack(loadSceneComplete)
	MapLoading.Instance:StartLoad(scene_id, old_scene_type)

	AudioManager.Instance:PlayEffect(ResPath.GetAudioEffectResPath(AudioEffect.GoCity),0.2)
end

function Scene:Update(now_time, elapse_time)
	self.is_in_update = true

	if nil ~= self.scene_logic then
		self.scene_logic:Update(now_time, elapse_time)
	end

	for _, v in pairs(self.obj_list) do
		if v ~= self.main_role then
			v:Update(now_time, elapse_time)
		end	
	end

	self.is_in_update = false

	-- 调用延时函数
	local delay_funcs = self.delay_handle_funcs
	self.delay_handle_funcs = {}
	for _, v in pairs(delay_funcs) do
		v()
	end

	self.main_role:Update(now_time, elapse_time)
	
	--更新场景特效
	for i = #self.effect_list, 1 , -1 do
		if self.effect_list[i]:IsTimeout() then
			local obj = table.remove(self.effect_list,i)
			obj:DeleteMe()
		else
			self.effect_list[i]:Update(now_time, elapse_time)
		end	
	end	

	--[[

	local main_role_logic_pos_x, main_role_logic_pos_y = self.main_role:GetLogicPos()
	local pos_change = false
	if main_role_logic_pos_x ~= 0 and main_role_logic_pos_y ~= 0 and 
		(self.cur_main_role_logic_pos.x ~= main_role_logic_pos_x or self.cur_main_role_logic_pos.y ~= main_role_logic_pos_y) then
		
		self.cur_main_role_logic_pos.x = main_role_logic_pos_x
		self.cur_main_role_logic_pos.y =main_role_logic_pos_y
		pos_change = true
	end

	if self.camera_end_scale ~= HandleGameMapHandler:GetCameraScale() then
		local cur_camer_scale = HandleGameMapHandler:GetCameraScale() + self.camera_scale_speed * elapse_time
		if self.camera_scale_speed > 0 and cur_camer_scale > self.camera_end_scale then
			cur_camer_scale = self.camera_end_scale
		elseif self.camera_scale_speed < 0 and cur_camer_scale < self.camera_end_scale then
			cur_camer_scale = self.camera_end_scale
		end
		HandleGameMapHandler:SetCameraScale(cur_camer_scale)
		pos_change = true
	end

	if pos_change then
		self:CheckClientObj()
	end
	--]]
end

function Scene:GetMainRole()
	return self.main_role
end

function Scene:CreateMainRole()
	if self.main_role then
		for k, v in pairs(self.obj_list) do
			if v:IsMainRole() then
				self.obj_list[k] = nil
				break
			end
		end

		self.main_role:DeleteMe()
		self.main_role = nil
	end

	local vo = GameVoManager.Instance:GetMainRoleVo()
	if nil == vo then
		Log("Scene:CreateMainRole vo nil")
		return nil
	end

	self.main_role = self:CreateObj(vo, SceneObjType.MainRole)
	if self.main_role then
		self.main_role:SetIsPingbiChibang(self.is_pinbi_wing)
		--self.main_role:SetIsPingbiDouli(self.is_pinbi_douli)
		self.main_role:SetTitleLayerVisible(not self:IsPingbiTitle())
		self.main_role:SetNameLayerSimple(self.is_simple_name)
	end
	SettingData.Instance:CalAutoSkill(false) --初始化挂机技能
	return self.main_role
end

function Scene:ChangeMainRoleObjId(obj_id)
	self.obj_list[self.main_role:GetObjId()] = nil
	self.main_role:GetVo().obj_id = obj_id
	self.obj_list[obj_id] = self.main_role
end

function Scene:CreateRole(role_vo)
	local role = self:CreateObj(role_vo, SceneObjType.Role)
	if nil ~= role then
		self.total_role_count = self.total_role_count + 1
		self:RefreshPingBiRoleOne(role)
		if MainRole.CacheAttackRole_OBJID == role_vo.obj_id then
			if not role_vo:GetIsHideState() then
				self.main_role:DoSelectSceneObj(role,false)
			end	
		end
	end
	return role
end

function Scene:CreateMonster(monster_vo)
	local monster = self:CreateObj(monster_vo, SceneObjType.Monster)
	monster:SetNameLayerShow(self.is_show_monster_name)
	return monster
end

function Scene:CreateFallItem(fallitem_vo)
	local fall_item = self:CreateObj(fallitem_vo, SceneObjType.FallItem)
	fall_item:SetNameLayerVisible(not self.is_no_fall_name)
	return fall_item
end

function Scene:CreateGather(gather_vo)
	return self:CreateObj(gather_vo, SceneObjType.GatherObj)
end

function Scene:CreateNpc(npc_vo)
	return self:CreateObj(npc_vo, SceneObjType.Npc)
end

function Scene:CreateDecoration(decoration_vo)
	return self:CreateObj(decoration_vo, SceneObjType.Decoration)
end



function Scene:CreateSpecialObj(special_vo)
	return self:CreateObj(special_vo, SceneObjType.SpecialObj)
end

local client_obj_id = -1
function Scene:CreateObj(vo, obj_type)
	if self.obj_list[vo.obj_id] then
		return nil
	end

	if 0 == vo.obj_id then
		vo.obj_id = client_obj_id
		client_obj_id = client_obj_id - 1
	end

	local obj = nil
	if obj_type == SceneObjType.Monster then
		obj = Monster.New(vo)
	elseif obj_type == SceneObjType.Role then
		obj = Role.New(vo)
	elseif obj_type == SceneObjType.MainRole then
		obj = MainRole.New(vo)
	elseif obj_type == SceneObjType.FallItem then
		local item_data  = Scene.Instance:GetFallItemScene()
		vo.effect_temp = {}
		if item_data[vo.item_id] then		
			vo.effect_temp = item_data[vo.item_id]
		end
		obj = FallItem.New(vo)
	elseif obj_type == SceneObjType.Npc then
		obj = Npc.New(vo)
	elseif obj_type == SceneObjType.Decoration then
		obj = Decoration.New(vo)
	elseif obj_type == SceneObjType.EffectObj then
		--obj = EffectObj.New(vo)
		ErrorLog("不能通过SceneCreateObj创建,管理不方便", obj_type)
		return
	elseif obj_type == SceneObjType.SpecialObj then
		obj = SpecialObj.New(vo)
	elseif	obj_type == SceneObjType.Hero then
		obj = Hero.New(vo)
	elseif	obj_type == SceneObjType.Gather then	
		obj = Gather.New(vo)
	elseif	obj_type == SceneObjType.Totem then
		obj = Totem.New(vo)
	elseif 	obj_type == SceneObjType.Humanoid then
		obj = HumanMonster.New(vo)
	elseif 	obj_type == SceneObjType.Pet then
		obj = Pet.New(vo)
	elseif obj_type == SceneObjType.DisplayMonster then	
		obj = DisplayMonster.New(vo)
	elseif 	obj_type == SceneObjType.Robot then	
		obj = Robot.New(vo)
	else 	
		obj_type = SceneObjType.Monster
		obj = Monster.New(vo) --默认创建都是怪物
	end

	if nil == obj then
		ErrorLog("Unknow obj type", obj_type)
		return
	end

	obj:Init(self)

	self.obj_list[obj:GetObjId()] = obj

	if 0 ~= obj:GetObjKey() then
		if nil == self.obj_group_list[obj_type] then
			self.obj_group_list[obj_type] = {}
		end
		self.obj_group_list[obj_type][obj:GetObjKey()] = obj
	end

	GlobalEventSystem:Fire(ObjectEventType.OBJ_CREATE, obj)
	if Story.Instance:GetIsStoring() and obj_type ~= SceneObjType.MainRole then
		obj:GetModel():SetAllVisible(false)
	end

	return obj
end

function Scene:DelObjHelper(obj_id)
	local del_obj = self.obj_list[obj_id]
	if del_obj then
		if del_obj == self.main_role then
			Log("Scene:DeleteObj warning: try to remove mainrole")
			return
		end

		if del_obj:GetType() == SceneObjType.Role then
			self.total_role_count = self.total_role_count - 1
			if self.total_role_count < 0 then self.total_role_count = 0 end
		end
		--print("移除对象:",del_obj.vo.name,del_obj:GetObjKey())
		self.obj_list[obj_id] = nil

		if nil ~= self.obj_group_list[del_obj:GetType()] then
			--print("移除组对象:",del_obj:GetType(),del_obj.vo.name,del_obj:GetObjKey())
			self.obj_group_list[del_obj:GetType()][del_obj:GetObjKey()] = nil
		end
		
		del_obj:DeleteMe()
		GlobalEventSystem:Fire(ObjectEventType.OBJ_DELETE, del_obj)
	end
end

function Scene:DeleteObj(obj_id)
	if self.is_in_update then
		-- update过程延迟删除
		if self.obj_list[obj_id] then
			table.insert(self.delay_handle_funcs, BindTool.Bind(self.DelObjHelper, self, obj_id))
		end
	else
		-- 直接删除对象
		self:DelObjHelper(obj_id)
	end
end

function Scene:DeleteObjByTypeAndKey(obj_type, obj_key)
	if nil ~= self.obj_group_list[obj_type] then
		local obj = self.obj_group_list[obj_type][obj_key]
		if nil ~= obj then
			self:DeleteObj(obj:GetObjId())
		end
	end
end

function Scene:GetObjectByObjId(obj_id)
	return self.obj_list[obj_id]
end

function Scene:GetObjectByTypeAndKey(obj_type, obj_key)
	if nil ~= self.obj_group_list[obj_type] then
		return self.obj_group_list[obj_type][obj_key]
	end
	return nil
end

function Scene:GetObjectByRoleId(role_id)
	local role_list = self:GetRoleList()
	for _,v in pairs(role_list) do
		if v.vo.role_id == role_id then
			return v
		end
	end
	return nil
end

function Scene:GetRoleByObjId(obj_id)
	local obj = self.obj_list[obj_id]
	if nil ~= obj and obj:IsRole() then
		return obj
	end

	return nil
end

function Scene:InitClientObj()
	if not self.scene_config then
		return
	end
end

function Scene.CalcSceneIndex(obj_type, cfg_index)
	return obj_type * 100000 + cfg_index
end

-- 是否工会成员
function Scene:IsGuildMember(target_obj)
	return RoleData.Instance:GetAttr(OBJ_ATTR.ACTOR_GUILD_ID) == target_obj:GetVo()[OBJ_ATTR.ACTOR_GUILD_ID] and target_obj:GetVo()[OBJ_ATTR.ACTOR_GUILD_ID] ~= 0
end

-- 是否友方
function Scene:IsFriend(target_obj)
	return self.scene_logic:IsFriend(target_obj, self.main_role)
end

-- 是否敌方
function Scene:IsEnemy(target_obj)
	return self.scene_logic:IsEnemy(target_obj, self.main_role)
end

-- 是否敌方
function Scene:IsCanSelecetByMode(target_obj)
	return self.scene_logic:IsCanSelecetByMode(target_obj, self.main_role)
end

--是否可攻击英雄
function Scene:IsCanSelectHero(target_obj)
	return self.scene_logic:IsCanSelectHero(target_obj, self.main_role)
end	

--选取对应坐标的对象数量
function Scene:CheckObjCountByCoordHelper(obj_list_group,x,y)
	local count = 0
	for _,obj_list in pairs(obj_list_group) do
		for _, v in pairs(obj_list) do
			if v:CanClick() then 
				local target_x, target_y = v:GetLogicPos()
				if target_x == x and target_y == y then
					count = count + 1
				end	
			end	
		end
	end		
	return count
end	

--一组坐标一次遍历记录要选择的对象数量
function Scene:CheckObjCountBySelfCoordsHelper(obj_list_group,low_level)
	low_level = low_level or 10000

	local count_list = {}
	for _,obj_list in pairs(obj_list_group) do
		for _, v in pairs(obj_list) do
			if v:CanClick() and v.vo[OBJ_ATTR.CREATURE_LEVEL] < low_level then 
				local target_x, target_y = v:GetLogicPos()
				local key = target_y * 10000 + target_x
				if count_list[key] then
					count_list[key] = count_list[key] + 1
				else
					count_list[key] = 1
				end	
			end
		end
	end		
	return count_list
end	

function Scene:CheckFireWallByPos(x, y)
	local list = self:GetSpecialObjs()
	for _,v in pairs(list) do
		if v:IsFireWall() then
			local target_x, target_y = v:GetLogicPos()
			if target_x == x and target_y == y then
				return true
			end	
		end	
	end
	return false	
end	

-- 选取最近的对象
function Scene:SelectObjHelper(obj_list_group, x, y, distance_limit)
	local target_obj = nil
	local target_distance = distance_limit
	for _,obj_list in pairs(obj_list_group) do
		for _, v in pairs(obj_list) do
			--过滤掉107场景属性的采集怪
			if v:CanClick() and not v:IsGatherAreaBox() then 
				local target_x, target_y = v:GetLogicPos()
				local distance = math.max(math.abs(target_x - x),math.abs(target_y - y))
				if distance < target_distance then
					target_obj = v
					target_distance = distance
				end
				
			end
		end
	end

	return target_obj, target_distance
end

--选择指定id的最近的怪物
function Scene:SelectMinDisMonster(monster_id, distance_limit)
	local target_obj = nil
	local target_distance = distance_limit
	target_distance = target_distance or 50
	local main_role = Scene.Instance:GetMainRole()
	local main_role_x, main_role_y = main_role:GetLogicPos()
	for _,obj_list in pairs(self:GetMonsterGroup()) do
		for _, v in pairs(obj_list) do
			if v:GetVo().monster_id == monster_id and v:CanClick() then
				local target_x, target_y = v:GetLogicPos()
				local distance = math.max(math.abs(target_x - main_role_x),math.abs(target_y - main_role_y))
				if distance < target_distance then
					target_obj = v
					target_distance = distance
				end
			end
		end
	end
	return target_obj
end

--获取最近的Boss
function Scene:GetMinDisBoss()
	local boss_id, x, y = nil, 0, 0
	local target_distance = 0
	local main_role = Scene.Instance:GetMainRole()
	local main_role_x, main_role_y = main_role:GetLogicPos()

	for _, v in pairs(BossData.Instance:GetOneSceneBossList(self:GetSceneId())) do
		target_x, target_y = MapData.GetMapBossPos(self:GetSceneId(), v)
		local distance = math.max(math.abs(target_x - main_role_x),math.abs(target_y - main_role_y))
		if boss_id == nil or distance < target_distance then
			boss_id = v
			x, y = target_x, target_y
			target_distance = distance
		end
	end
	return boss_id, x, y
end



--选择最近的可拾取掉落
function Scene:SelectMinRemindFallItem(distance_limit)
	local target_obj = nil
	local target_distance = distance_limit
	target_distance = target_distance or 50
	local main_role = Scene.Instance:GetMainRole()
	local main_role_x, main_role_y = main_role:GetLogicPos()
	local obj_id = nil

	local item_t = self:GetFallItemList()

	for _, v in pairs(item_t) do
		local target_x, target_y = v:GetLogicPos()
		if SettingProtectData.CheckPickFallItemSetting(v.vo) 
			and not GameMapHelper.IsBlock(target_x, target_y) 
			and not GameMapHelper.IsDynamicNotMoveCoord(target_x, target_y) 
			and not v.vo:GetIsLock() then
			obj_id = v:GetObjId()
			local flat = self:FallItemHasPick(obj_id)
			if flat ~= true and not self.auto_fight_record_not_pick[obj_id] then
				
				local distance = math.abs(target_x - main_role_x) + math.abs(target_y - main_role_y)
				if distance < target_distance then
					target_obj = v
					target_distance = distance
				end
			end
		end
	end
	return target_obj
end

function Scene:RecordFightNotPick(obj_id)
	self.auto_fight_record_not_pick[obj_id] = true
end	

function Scene:ClearFightNotPick()
	self.auto_fight_record_not_pick = {}
end	

function Scene:RecordHasItemPick(obj_id)
	local flat = self.has_pick_fallitem_t[obj_id]
	if flat == nil then
		self.has_pick_fallitem_t[obj_id] = true
		self.has_pick_count = self.has_pick_count + 1
		self:CheckRealeaseItemPick() 
	end	
end	

function Scene:CheckRealeaseItemPick()
	--记录超过100个
	if self.has_pick_count >= 100 then
		local new_record_t = {}
		local has_count = 0
		local item_t = self:GetFallItemList()
		for _, v in pairs(item_t) do
			if self.has_pick_fallitem_t[v:GetObjId()] then --当前物品存在的继续保持
				new_record_t[v:GetObjId()] = true
				has_count = has_count + 1
			end	
		end	
		self.has_pick_count = has_count -- 释放原来的表
		self.has_pick_fallitem_t = new_record_t
	end	
end	

function Scene:FallItemHasPick(obj_id)
	return self.has_pick_fallitem_t[obj_id]
end


function Scene:GetNpcByNpcId(npc_id)
	for _, v in pairs(self:GetObjListByType(SceneObjType.Npc)) do
		if v:GetNpcId() == npc_id then
			return v
		end
	end
	return nil
end

local empty_table = {}
function Scene:GetObjListByType(obj_type)
	return self.obj_group_list[obj_type] or empty_table
end

function Scene:GetRoleList()
	return self.obj_group_list[SceneObjType.Role] or empty_table
end

function Scene:GetHeroList()
	return self.obj_group_list[SceneObjType.Hero] or empty_table
end

function Scene:GetMonsterList()
	return self.obj_group_list[SceneObjType.Monster] or empty_table
end

function Scene:GetNpcList()
	return self.obj_group_list[SceneObjType.Npc] or empty_table
end

function Scene:GetGatherList()
	return self.obj_group_list[SceneObjType.Gather] or empty_table
end

function Scene:GetFallItemList()
	return self.obj_group_list[SceneObjType.FallItem] or empty_table
end

function Scene:GetTotemList()
	return self.obj_group_list[SceneObjType.Totem] or empty_table
end	

function Scene:GetHumanoidList()
	return self.obj_group_list[SceneObjType.Humanoid] or empty_table
end	

function Scene:GetPetList()
	return self.obj_group_list[SceneObjType.Pet] or empty_table
end	

function Scene:GetMonsterGroup()
	return {self:GetMonsterList(),self:GetGatherList(),self:GetTotemList(),self:GetHumanoidList()}
end	

function Scene:GetRobot()
	return self.obj_group_list[SceneObjType.Robot] or empty_table
end	

function Scene:GetSpecialObjs()
	return self.obj_group_list[SceneObjType.SpecialObj] or empty_table
end	

function Scene:OnSceneObjPosChange(x, y, sender)
	if self.is_check_all_hide then
		if self.main_role 
			and sender ~= self.main_role 
			and sender:GetType() == SceneObjType.Role then
			local logic_x = self.main_role.logic_pos.x
			local logic_y = self.main_role.logic_pos.y
			

			if HandleGameMapHandler:GetGameMap() then
				local zone_info = HandleGameMapHandler:GetGameMap():getZoneInfo(x, y) or 0
				--print("玩家移动:",x,y,zone_info)
				if zone_info == ZoneType.AllHide then
					sender:SetAllHide(true)
				else	
					local dis = math.max(math.abs(x - logic_x),math.abs(y - logic_y))
					if dis >= self.check_all_hide_range then
						sender:SetAllHide(true)
					else
						sender:SetAllHide(false)	
					end	
				end	
			else
				local dis = math.max(math.abs(x - logic_x),math.abs(y - logic_y))
				if dis >= self.check_all_hide_range then
					sender:SetAllHide(true)
				else
					sender:SetAllHide(false)	
				end	
			end

		end	
	end
end

function Scene:OnRoleAttrChange(key, value, old_value)
	if self.is_check_all_hide and self.client_scene_cfg then
		if key == OBJ_ATTR.ACTOR_CAMP then
			self:SetHideRange()
		end	
	end	
end	

function Scene:ItemDataChangeCallback(change_type, item_id)
	-- if item_id == 111 then
	-- end	
	if self.is_check_all_hide and self.client_scene_cfg then
		local items = self.client_scene_cfg.checkItems
		if items then
			local is_check = false
			for _,v in pairs(items) do
				if v == item_id then
					is_check = true
					break
				end	
			end	

			if is_check then
				self:SetHideRange()
			end
		end	
	end	
end	

function Scene:SetHideRange()
	local range = self.client_scene_cfg.defaultRange or 0
	local items = self.client_scene_cfg.checkItems
	if items then
		for i = #items , 1, -1 do
			local num = ItemData.Instance:GetItemNumInBagById(items[i])
			if num > 0 then
				range = self.client_scene_cfg.itemRanges[i]
				break	
			end	
		end	
	end	
	local camp = RoleData.Instance:GetAttr(OBJ_ATTR.ACTOR_CAMP) or 0
	if self.client_scene_cfg.campRange then
		range = math.max(range,self.client_scene_cfg.campRange[camp] or 0)
	end	
	
	self.check_all_hide_range = range
	local x,y = self.main_role:GetLogicPos()
	self:OnMainRolePosChange(x,y)

	GlobalEventSystem:Fire(SceneEventType.SCENE_ROLE_RANGE_CHANGE)
end	


function Scene:OnMainRolePosChange(x, y)
	if self.is_check_all_hide then
		local dis = 0
		local pos = nil
		local role_list = self:GetRoleList()
		for _, v in pairs(role_list) do
			pos = v.logic_pos
			if HandleGameMapHandler:GetGameMap() then
				local zone_info = HandleGameMapHandler:GetGameMap():getZoneInfo(pos.x, pos.y) or 0
				--print("用户移动:",pos.x, pos.y,zone_info)
				if zone_info == ZoneType.AllHide then
					v:SetAllHide(true)
				else	
					dis = math.max(math.abs(pos.x - x),math.abs(pos.y - y))
					if dis >= self.check_all_hide_range then
						v:SetAllHide(true)
					else
						v:SetAllHide(false)	
					end	
				end	
			else
				dis = math.max(math.abs(pos.x - x),math.abs(pos.y - y))
				if dis >= self.check_all_hide_range then
					v:SetAllHide(true)
				else
					v:SetAllHide(false)	
				end	
			end	
				
		end
	end
end

function Scene:DrawEatChickenArea(indexs)
	if self.eat_chicken_range_list then
		self.draw_node:clear()
		for _,v in pairs(indexs) do
			local range = self.eat_chicken_range_list[v]
			if range  then
				local vec_list = {}
				local min_x = range[1]
				local min_y = range[2]
				
				local max_x = range[3] + 1
				local max_y = range[4] + 1
				
				vec_list[1] = cc.p(min_x * Config.SCENE_TILE_WIDTH, min_y * Config.SCENE_TILE_HEIGHT)
				vec_list[2] = cc.p(max_x * Config.SCENE_TILE_WIDTH, min_y * Config.SCENE_TILE_HEIGHT)
				vec_list[3] = cc.p(max_x * Config.SCENE_TILE_WIDTH, max_y * Config.SCENE_TILE_HEIGHT)
				vec_list[4] = cc.p(min_x * Config.SCENE_TILE_WIDTH, max_y * Config.SCENE_TILE_HEIGHT)
				self.draw_node:drawPolygon(vec_list ,#vec_list ,cc.c4f(1, 0, 0, 0.2) , 0.01 , cc.c4f(0, 0, 0, 0.01))
			end	
		end	
	end	
end	


function Scene:OnMainRoleMoveStart()

end

function Scene:OnMainRoleMoveEnd()

end

function Scene:OnMainRoleDead(main_role)
	main_role:StopMove()
end

function Scene:CheckClientObj()
	if not self.scene_config then
		return
	end
	--[[
	local rect = HandleRenderUnit:GetCoreScene():GetViewRect()
	for k, v in pairs(self.scene_config.decorations) do
		local x, y = HandleRenderUnit:LogicToWorldXY(v.x, v.y)
		x = x * HandleGameMapHandler:GetCameraScale()
		y = y * HandleGameMapHandler:GetCameraScale()
		
		if GameMath.IsInRect(x, y, rect.x, rect.y, rect.width, rect.height) then
			if nil == self:GetObjectByTypeAndKey(SceneObjType.Decoration, k) then
				local decoration_vo = GameVoManager.Instance:CreateVo(DecorationVo)
				decoration_vo.obj_key = k
				decoration_vo.decoration_id = v.id
				decoration_vo.name = v.name
				decoration_vo.pos_x = v.x
				decoration_vo.pos_y = v.y
				local decoration = self:CreateDecoration(decoration_vo)
			end
		else
			self:DeleteObjByTypeAndKey(SceneObjType.Decoration, k)
		end
	end
	--]]
end



----------------------------------------------------
-- 在场景上生成一次性特效
----------------------------------------------------
function Scene:CreateOnceEffect(effect_id, pos_x, pos_y, layer, end_callback)
	local effect = nil
	function callback_func()
		if end_callback ~= nil then
			end_callback()
		end
		NodeCleaner.Instance:AddNode(effect)
	end

	local anim_path, anim_name = ResPath.GetEffectAnimPath(effect_id)
	effect = RenderUnit.CreateAnimSprite(anim_path, anim_name, FrameTime.Effect, 1, false, callback_func)
	effect:setPosition(pos_x, pos_y)
	HandleRenderUnit:GetCoreScene():addChildToRenderGroup(effect, layer or GRQ_SCENE_OBJ)

	return effect
end

function Scene.PlayOneFlyEffect(effect_id, time, pos_x, pos_y, layer, end_callback)
	local screen_w = HandleRenderUnit:GetWidth()
	local screen_h = HandleRenderUnit:GetHeight()
	local screen_w_m = screen_w * 0.5
	local screen_h_m = screen_h * 0.5
	pos_x = pos_x or screen_w_m
	pos_y = pos_y or screen_h_m

	local anim_path, anim_name = ResPath.GetEffectUiAnimPath(effect_id or 918)
	local fly_effect = AnimateSprite:create(anim_path, anim_name, 99, FrameTime.Effect, false)
	fly_effect:setPosition(pos_x, pos_y)
	HandleRenderUnit:AddUi(fly_effect, layer or COMMON_CONSTS.ZORDER_SYSTEM_EFFECT)

	-- local bezier_to = cc.BezierTo:create(time or 1.4, {cc.p(screen_w * math.random(), screen_h * 0.9), cc.p(screen_w * math.random(), pos_y), cc.p(screen_w_m, 3)})
	local bezier_to = cc.BezierTo:create(time or 1.4, {cc.p(screen_w * math.random(), screen_h * math.random()), cc.p(screen_w * math.random(), pos_y), cc.p(screen_w_m, 3)})
	local callback = cc.CallFunc:create(function ()
		if end_callback ~= nil then
			end_callback()
		end
		fly_effect:setStop()
		NodeCleaner.Instance:AddNode(fly_effect)
	end)
	local action = cc.Sequence:create(bezier_to, callback)
	fly_effect:runAction(action)
end


-- 刷新单个屏蔽玩家
function Scene:RefreshPingBiRoleOne(role)
	if role:IsMainRole() then
		return
	end

	if self:GetIsInAppearLimitList(role) then
		return
	end

	local is_pingbi = self:IsNeedPingBiRole(role)
	local role_v = not is_pingbi
	
	self:SetRoleReleateVisible(role, role_v)
end

-- 刷新屏蔽宠物
function Scene:RefreshPingBiPet()
	local is_visible = not self:IsPingbiPet()
	for k, v in pairs(self:GetPetList()) do
		v:SetIsNotMaskModel(is_visible)
	end
end

-- 刷新屏蔽战将
function Scene:RefreshPingBiHero()
	local is_visible = not self:IsPingbiHero()

	for k, v in pairs(self:GetObjListByType(SceneObjType.Hero)) do
		v:SetIsNotMaskModel(is_visible)
	end
end

-- 刷新缩小宠物
function Scene:RefreshLittlePet()
	for k, v in pairs(self:GetPetList()) do
		v:GetModel():SetScale(self.is_little_pet and 0.66 or 1)
	end
end

-- 刷新屏蔽怪
function Scene:RefreshPingBiMonster()
	if nil ~= self.obj_group_list[SceneObjType.Monster] then
		for k, v in pairs(self.obj_group_list[SceneObjType.Monster]) do
			if v:IsCommon() then
				v:SetIsNotMaskModel(not self.is_pinbi_monster)
			end
		end
	end
end

-- 刷新屏蔽角色
function Scene:RefreshPingBiRole()
	if nil ~= self.obj_group_list[SceneObjType.Role] then
		for k, v in pairs(self.obj_group_list[SceneObjType.Role]) do
			self:RefreshPingBiRoleOne(v)
		end
	end
	local robots = self:GetRobot()
	for k, v in pairs(robots) do
		v:SetIsNotMaskModel(not self.is_pingbi_other_role)	
	end	
end


--设置角色相关的可见性
function Scene:SetRoleReleateVisible(role, role_v)
	if role == nil or 1 == role:GetVo().is_shadow then return end
	role:SetIsNotMaskModel(role_v)
	role:SetIsPingbiChibang(role_v and self.is_pinbi_wing)
--	role:SetIsPingbiDouli(role_v and self.is_pinbi_douli)
	role:SetTitleLayerVisible(not self:IsPingbiTitle())
	role:SetNameLayerSimple(self.is_simple_name)
end

function Scene:SetRoleAllReleateVisible(role, visible)
	if nil == role or 1 == role:GetVo().is_shadow then return end
	self:SetRoleReleateVisible(role, visible)
end

-- 是否是要屏蔽的角色
function Scene:IsNeedPingBiRole(role)
	if 1 == role:GetVo().is_shadow then
		 return false
	end
	local is_pingbi = false

	if self.is_pingbi_other_role then
		is_pingbi = true
	else
		if self.is_pingbi_guild_role and self:IsGuildMember(role) then
			is_pingbi = true
		end
	end
	return is_pingbi
end

-- 系统设置改变
function Scene:OnSysSettingChange(setting_type, flag)
	if setting_type == SETTING_TYPE.SHIELD_OTHERS then
		self.is_pingbi_other_role = flag
		self:RefreshPingBiRole()
	elseif setting_type == SETTING_TYPE.SHIELD_SAME_CAMP then
		self.is_pingbi_guild_role = flag
		self:RefreshPingBiRole()
	elseif setting_type == SETTING_TYPE.CLOSE_TITLE then
		self.is_no_title = flag
		self:RefreshTitleView(self.is_no_title)
	elseif setting_type == SETTING_TYPE.SIMPLE_ROLE_NAME then
		self.is_simple_name = flag
		self:RefreshNameView(self.is_simple_name)
	elseif setting_type == SETTING_TYPE.SHIELD_FALL_NAME then
		self.is_no_fall_name = flag
		self:RefreshFallNameView(self.is_no_fall_name)
	elseif setting_type == SETTING_TYPE.SHIELD_MONSTER_NAME then
		self.is_show_monster_name = flag
		self:RefreshMonsterNameView(self.is_show_monster_name)
	elseif setting_type == SETTING_TYPE.SHIELD_MONSTER then
		self.is_pinbi_monster = flag
		self:RefreshPingBiMonster()
	elseif setting_type == SETTING_TYPE.SHIELD_PET then
		self.is_pinbi_pet = flag
		self:RefreshPingBiPet()
	elseif setting_type == SETTING_TYPE.SHIELD_WING then
		self.is_pinbi_wing = flag
		self:RefreshPingBiWing(self.is_pinbi_wing)
	elseif setting_type == SETTING_TYPE.SHIELD_HERO_PET_NAME then
		--self.is_pinbi_douli = flag
		--self:RefreshPingBiDouli(self.is_pinbi_douli)
		self.is_show_pet_hero_name = flag
		self:RefreshHeroPetNameView(self.is_show_pet_hero_name)
	elseif setting_type == SETTING_TYPE.LITTLE_SHENSHOU then
		self.is_little_pet = flag
		self:RefreshLittlePet(self.is_little_pet)
	elseif setting_type == SETTING_TYPE.SHIELD_HERO then	
		self.is_pinbi_hero = flag
		self:RefreshPingBiHero()	
	end
end

-- 挂机设置改变
function Scene:OnGuaJiSettingChange(guaji_setting_type, flag)
	if GUAJI_SETTING_TYPE.GUAJI_PICKUP == guaji_setting_type then
		self.pick_color_flag = flag
	end
end

-- 帧率回调
function Scene:OnFpsCallback(fps)
	if not self.is_check_fps or fps <= 0 then
		return
	end	

	self.total_fps = self.total_fps + fps
	self.fps_tick = self.fps_tick + 1

	local now_time = Status.NowTime
	if now_time > self.fps_time then --间隔3秒检测
		self.fps_time = now_time + 3

		local old_fps = self.fps_now
		self.fps_now = self.total_fps / self.fps_tick

		if self.fps_now <= 30 then
		local appear_count = math.floor(SceneAppearRoleCount.Max * (1 - (30 - self.fps_now) * 0.1))
			self:SetMaxAppearRoleCount(appear_count)
		end

		if (old_fps >= fps_title and self.fps_now < fps_title) or (old_fps < fps_title and self.fps_now >= fps_title) then
			self:RefreshTitleView()
		end
		if (old_fps >= fps_wing and self.fps_now < fps_wing) or (old_fps < fps_wing and self.fps_now >= fps_wing) then
			self:RefreshPingBiWing()
		end
		if (old_fps >= fps_douli and self.fps_now < fps_douli) or (old_fps < fps_douli and self.fps_now >= fps_douli) then
			self:RefreshPingBiDouli()
		end
		if (old_fps >= fps_hero and self.fps_now < fps_hero) or (old_fps < fps_hero and self.fps_now >= fps_hero) then
			self:RefreshPingBiHero()
		end
		if (old_fps >= fps_pet and self.fps_now < fps_pet) or (old_fps < fps_pet and self.fps_now >= fps_pet) then
			self:RefreshPingBiPet()
		end

		--print("平均帧数:",old_fps,self.fps_now)

		self.total_fps = 0
		self.fps_tick = 0

	end	
end


function Scene:IsPingbiTitle()
	return self.is_no_title or (self.fps_now < fps_title) or ClientMaskInfoCfg.MaskRoleChangeMonsterNameMap[self:GetSceneId()]
end

function Scene:IsPingbiWing()
	return self.is_pinbi_wing or (self.fps_now < fps_wing)
end

function Scene:IsPingbiDouli()
	return self.is_pinbi_douli or (self.fps_now < fps_douli)
end

-- 是否屏蔽战将
function Scene:IsPingbiHero()
	return self.is_pinbi_hero or self.fps_now < fps_hero
end

-- 是否屏蔽宠物
function Scene:IsPingbiPet()
	return self.is_pinbi_pet or self.fps_now < fps_pet
end

----------------------------------------------------
-- 角色个数控制
----------------------------------------------------
function Scene:IsNeedLimitAppear()
	return self.total_role_count > self.max_appear_role_count
end

function Scene:AddToAppearLimitList(role_obj)
	if role_obj == nil or role_obj:GetType() ~= SceneObjType.Role then return end

	self.limit_appear_role_list[role_obj:GetObjId()] = 1
	self:SetRoleAllReleateVisible(role_obj, false)
end

function Scene:RemoveFromAppearLimitList(role_obj)
	if role_obj == nil or role_obj:GetType() ~= SceneObjType.Role then return end

	self.limit_appear_role_list[role_obj:GetObjId()] = nil
	self:SetRoleAllReleateVisible(role_obj, true)
	self:RefreshPingBiRoleOne(role_obj)
end

function Scene:GetIsInAppearLimitList(role_obj)
	if role_obj == nil then return end
	return nil ~= self.limit_appear_role_list[role_obj:GetObjId()]
end

function Scene:RefreshAppearLimitList()
	for k,v in pairs(self.limit_appear_role_list) do
		local role_obj = self:GetObjectByObjId(k)
		if role_obj and role_obj:GetType() == SceneObjType.Role then
			self:SetRoleAllReleateVisible(role_obj, false)
		end
	end
end

function Scene:RefreshTitleView(no_title_show)
	local is_visible = not self:IsPingbiTitle()
	self.main_role:SetTitleLayerVisible(is_visible)
	for k,v in pairs(self:GetObjListByType(SceneObjType.Role)) do
		v:SetTitleLayerVisible(is_visible)
	end
end

function Scene:RefreshPingBiWing(is_pinbi_wing)
	local is_pinbi = self:IsPingbiWing()
	self.main_role:SetIsPingbiChibang(is_pinbi)
	for k,v in pairs(self:GetObjListByType(SceneObjType.Role)) do
		v:SetIsPingbiChibang(is_pinbi)
	end
	local robots = self:GetRobot()
	for k,v in pairs(robots) do
		v:SetIsPingbiChibang(is_pinbi)
	end
end

function Scene:RefreshPingBiDouli(is_pinbi_douli)
	self.main_role:SetIsPingbiDouli(is_pinbi_douli)
	for k,v in pairs(self:GetObjListByType(SceneObjType.Role)) do
		v:SetIsPingbiDouli(is_pinbi_douli)
	end
end

function Scene:RefreshNameView(is_simple_name)
	self.main_role:SetNameLayerSimple(is_simple_name)
	for k,v in pairs(self:GetObjListByType(SceneObjType.Role)) do
		v:SetNameLayerSimple(is_simple_name)
	end
	-- local robots = self:GetRobot()
	-- for k,v in pairs(robots) do
	-- 	v:SetNameLayerShow(is_simple_name)
	-- end
end

function Scene:RefreshFallNameView(no_name_show)
	for k,v in pairs(self:GetObjListByType(SceneObjType.FallItem)) do
		v:SetNameLayerVisible(not no_name_show)
	end
end

function Scene:RefreshMonsterNameView(name_show)
	local group = self:GetMonsterGroup()
	for _,l in pairs(group) do
		for k,v in pairs(l) do
			v:SetNameLayerShow(name_show)
		end
	end
end

function Scene:RefreshHeroPetNameView(name_show)
	local group = {self:GetObjListByType(SceneObjType.Hero),self:GetObjListByType(SceneObjType.Pet)}
	for _,l in pairs(group) do
		for k,v in pairs(l) do
			v:SetNameLayerShow(name_show)
		end
	end
end

-- 设置同屏最大人数限制
function Scene:SetMaxAppearRoleCount(appear_count)
	local old_count = self.max_appear_role_count
	self.max_appear_role_count = math.min(appear_count, SceneAppearRoleCount.Max)
	self.max_appear_role_count = math.max(appear_count, SceneAppearRoleCount.Min)
	if self.max_appear_role_count >= old_count then
		return
	end

	-- 已屏蔽数量
	local scene_pingbi_count = 0
	for k, v in pairs(self.limit_appear_role_list) do
		scene_pingbi_count = scene_pingbi_count + 1
	end

	-- 还需要屏蔽数量
	local need_pingbi_count = (self.total_role_count - scene_pingbi_count) - self.max_appear_role_count
	if need_pingbi_count <= 0 then
		return
	end

	for k, v in pairs(self:GetObjListByType(SceneObjType.Role)) do
		if nil == self.limit_appear_role_list[k] then
			self:AddToAppearLimitList(v)

			need_pingbi_count = need_pingbi_count - 1
			if need_pingbi_count <= 0 then
				break
			end
		end
	end
end

-- 获取特效数量
-- @return 总数，相同数
function Scene:GetEffectCount(effect_id)
	local total_count, same_count = 0, 0

	for k, v in pairs(self:GetObjListByType(SceneObjType.EffectObj)) do
		total_count = total_count + 1
		if v:GetVo().effect_id == effect_id then
			same_count = same_count + 1
		end
	end

	return total_count, same_count
end

----------------------------------------------------
-- 场景缩放
----------------------------------------------------
function Scene:SetSceneCameraScaleTo(scale, use_time)
	scale = scale or 1
	use_time = use_time or 1
	if scale > 0 and scale ~= HandleGameMapHandler:GetCameraScale() then
		Log("场景缩放到:", use_time, scale)
		self.camera_end_scale = scale
		self.camera_scale_speed = (scale - HandleGameMapHandler:GetCameraScale()) / use_time
	end
end

--是否是 自动挂机不采集 区域
function Scene:IsAutoBattleNotGatherArea()
	return self.cur_area_info.attr_t[MapAreaAttribute.aaAutoBattleNotGather]
end	

--是否是pk区域
function Scene:IsPkArea()
	return self.cur_area_info.attr_t[MapAreaAttribute.aaSaft]
end	

--是否是禁止召唤
function Scene:IsNotZhaohuanArea()
	return self.cur_area_info.attr_t[MapAreaAttribute.aaNotCallHero]
end	

--是不爆装备区域
function Scene:IsNoDropEquipArea()
	return self.cur_area_info.attr_t[MapAreaAttribute.asNoDropEquip]
end	

--是不掉幸运区域
function Scene:IsNotDropLuck()
	return self.cur_area_info.attr_t[MapAreaAttribute.aaNotDropLuck]
end	

--禁止召唤战将
function Scene:IsNotZhaohuanZhanjiang()
	return self.cur_area_info.attr_t[MapAreaAttribute.aaNotCallZhanJiang]
end	

--是否pk加pk值
function Scene:IsPkValue()
	return self.cur_area_info.attr_t[MapAreaAttribute.aaWar]
end	

--是不能原地复活区域
function Scene:IsNotFuhuo()
	return self.cur_area_info.attr_t[MapAreaAttribute.aaNotHereRelive]
end	

--不能使用
function Scene:IsNotUse()
	return self.cur_area_info.attr_t[MapAreaAttribute.aaNotItemId]
end	